/*
 *   Copyright (c) 1993 by
 *   PARALLAX GRAPHICS, INCORPORATED, Santa Clara, California.
 *   All rights reserved
 *
 *   This software is furnished on an as-is basis, and may be used and copied
 *   only with the inclusion of the above copyright notice.
 *
 *   The information in this software is subject to change without notice.
 *   No committment is made as to the usability or reliability of this
 *   software.
 *
 *   Parallax Graphics, Inc.
 *   2500 Condensa Street
 *   Santa Clara, California  95051
 *   Author: George Drapeau
 */
/* The purpose of this program is to show the ability to digitize images
 * from the screen and save them as raw JPEG data on disk.  In this case,
 * the images are digitized from an incoming video signal, then compressed,
 * then saved to disk.
 * The program takes several arguments:
 * - the number of frames to be recorded and saved to disk
 * - the width and height of the image to be digitized and compressed
 * - the name of the disk file into which the JPEG data will be written.
 * The program creates a VideoIn widget and a JPEG widget, then informs the
 * JPEG widget that it should capture image data from the VideoIn widget (the
 * JPEG widget can point to any window, but in practice only 24-bit windows
 * provide the JPEG widget with workable data; for this example, it seemed
 * reasonable to use video as the best source of 24-bit data).
 * The program then sets up a function to begin work after a short delay.
 * This timer function does all of the work, doing the following:
 * - Before beginning work, it saves the current time-of-day for later
 *   statistics purposes.
 * - For each frame to be saved, the program tells the JPEG widget to digitize
 *   and compress a new frame, then writes that JPEG data to disk.
 * - After compressing and saving all the frames requested, the program
 *   gathers statistics to indicate performance of the widget, then reports
 *   these statistics before exiting.
 *
 * Usage: jpegTimingTest <numberOfFramesToBeSaved> <width> <height> <fileName>
 */
/* $Header: /plx/S/source/Xclients/solaris2.4/3.4.new/RCS/jpegTimingTest.c,v 1.11 1993/11/17 02:31:37 drapeau Exp $ */
/* $Log: jpegTimingTest.c,v $
 * Revision 1.11  1993/11/17  02:31:37  drapeau
 * No real modifications; only small cosmetic changes.
 *
 * Revision 1.10  93/09/10  12:37:35  drapeau
 * First public version of this test program.
 *  */
static char jpegTimingTestRcsId[] = "$Header: /plx/S/source/Xclients/solaris2.4/3.4.new/RCS/jpegTimingTest.c,v 1.11 1993/11/17 02:31:37 drapeau Exp $";

#include <stdio.h>
#include <sys/param.h>
#include <sys/time.h>
#include <signal.h>
#include <fcntl.h>
#include <X11/Xlib.h>
#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <XPlxExt.h>
#include <XVideoIn.h>
#include <JPEG.h>

Widget top, videoShell, jpegShell, videoWidget, jpegWidget;
void	TimerFunction(XtPointer, XtIntervalId*);

int	 	totalNumberOfGrabs = 10;
int		outputFile;
Boolean		storeCompressedDataToDisk;
char*		fileName;
XRectangle	myRect;

 
void main(int argc, char** argv)
{
  XtAppContext	appContext;
  XtIntervalId	timerID;
  
  top = XtAppInitialize(&appContext, "JPEGTest", NULL, 0,
			       &argc, argv,
			       NULL, (ArgList)NULL, 0);
  if (argc < 5)							    /* Were enough arguments supplied? */
  {								    /* No, print an error message and exit immediately */
    printf("Usage: %s <numberOfFramesToBeSaved> <width> <height> <fileName>.\n",
	   argv[0]);
    exit(-1);
  }
  totalNumberOfGrabs = (int)atoi(argv[1]);			    /* Determine how many frames to compress and save to disk */
  myRect.x = 0;
  myRect.y = 0;
  myRect.width = (unsigned short)atoi(argv[2]);			    /* Determine the dimensions of the source image */
  myRect.height = (unsigned short)atoi(argv[3]);
  fileName = argv[4];						    /* Determine file name into which to store compressed data */
  if (argc > 5)							    /* Was an option specified to skip writing data to disk? */
  {								    /* Yes, set the appropriate variable and report this */
    printf("Will not store compressed data to disk.\n");
    storeCompressedDataToDisk = False;
  }
  else								    /* No, normal operation: store data to disk */
  {
    storeCompressedDataToDisk = True;
    outputFile = open(fileName, (O_CREAT | O_WRONLY), 0644);	    /* Try to create a new file for writing data to disk */
    if (outputFile == -1)					    /* Did the open fail? */
    {								    /* Yes, report the error and exit this program without... */
      printf("Could not open the file %s for writing.  Exiting.\n", /* ...doing anything more. */
	     fileName);
      exit(outputFile);
    }
  }
  XtVaSetValues(top,						    /* Resize the top-level widget to fit the size of the... */
		XtNwidth, (Dimension)myRect.width,		    /* ...images to be digitized and compressed. */
		XtNheight, (Dimension)myRect.height,
		NULL);
  videoWidget = XtVaCreateManagedWidget("videoIn",		    /* Create a widget for displaying video */
					xVideoInWidgetClass, 
					top,
					NULL);
  XtRealizeWidget(top);
  jpegShell = XtVaCreateManagedWidget ("jpegShell",
				       applicationShellWidgetClass, 
				       top,
				       NULL);
  jpegWidget = XtVaCreateManagedWidget("jpeg", jpegWidgetClass,	    /* Create a widget for compressing JPEG image data */
				       jpegShell,
				       XtNcompressSourceRect,
				       (XtPointer)&myRect,
				       XtNcompressDestRect,
				       (XtPointer)&myRect,
				       XtNsourceWin,
				       XtWindow(videoWidget),
				       NULL);
  XtRealizeWidget(jpegShell);
  timerID = XtAppAddTimeOut(appContext, 2500,			    /* In a little while, start doing the actual work */
			    TimerFunction, (XtPointer)NULL);
  XtAppMainLoop(appContext);
}								    /* end function main */


void TimerFunction(XtPointer clientData, XtIntervalId* id)
{
  int			counter;
  int			result;
  int			grabsPerSecond;
  int			numberBytesWritten = 0;
  int			averageDiskBandwidth = 0;
  int			largestFrameSize = 0;
  int			smallestFrameSize = 0;
  int			qFactor;
  static Boolean	firstTime = True;
  Boolean		useSharedMemory;
  struct timeval	before;
  struct timeval	after;
  struct timeval	totalTime;
  struct timeval	timeForOneGrab;
  XPlxCImage*		jpegDataPtr;
  
  printf("Entered TimerFunction.\n\n\n");
  XtVaSetValues(jpegWidget,					    /* Tell the JPEG widget not to bother trying to draw the... */
		XtNdestWin, 0,					    /* ...images it just compressed.  This will save time and... */
		NULL);						    /* ...allow the program to write data to disk more quickly. */
  XtVaGetValues(jpegWidget,					    /* Determine if the JPEG widget is using shared memory... */
		XtNuseSharedMemory, &useSharedMemory,		    /* ...for speed purposes.  If so, point to the shared... */
		XtNjpegData, &jpegDataPtr,			    /* ...memory segment since it will be recycled. */
		XtNqFactor, &qFactor,				    /* Also get the current Q factor used by the JPEG widget */
		NULL);
  result = gettimeofday(&before, (struct timezone*)NULL);	    /* Determine the current time-of-day for statistics purposes */
  for (counter = 0; counter < totalNumberOfGrabs; counter++)	    /* Loop once through for each image to be saved */
  {
    XtVaSetValues(jpegWidget,					    /* Tell the JPEG widget to compress another image right now */
		  XtNgetNewImage, True,
		  NULL);
    if (useSharedMemory == False)				    /* Is the widget using shared memory for storing images? */
      XtVaGetValues(jpegWidget,					    /* No, explicitly point to the newest JPEG data (if the... */
		    XtNjpegData, &jpegDataPtr,			    /* ...widget isn't using shared memory, each new image... */
		    NULL);					    /* ...points to a different JPEG data structure). */
    if (storeCompressedDataToDisk == True)			    /* Is this test supposed to write JPEG data to disk? */
    {								    /* Yes, write one image, if possible, to disk. */
      result = write(outputFile, jpegDataPtr, sizeof(XPlxCImage));  /* Write the image's header first */
      result = write(outputFile, jpegDataPtr->data,		    /* Next, write the image data itself */
		     jpegDataPtr->size);
    }
    numberBytesWritten +=					    /* Update total number of bytes written so far */
      (sizeof(XPlxCImage) + jpegDataPtr->size);
    if (firstTime == True)					    /* Is this the first time through the compression loop? */
    {								    /* Yes, initialize the smallest frame size so far seen to... */
      smallestFrameSize = jpegDataPtr->size;			    /* ...be the first frame compressed. */
      firstTime = False;
    }
    if (jpegDataPtr->size > largestFrameSize)			    /* Update largest and smallest compressed frames seen...  */
      largestFrameSize = jpegDataPtr->size;			    /* ...so far. */
    if (jpegDataPtr->size < smallestFrameSize)
      smallestFrameSize = jpegDataPtr->size;
    if ((result < jpegDataPtr->size) &&
	(storeCompressedDataToDisk == True))
    {
      printf("There was an error in writing image number %d; only %d bytes were written when %d were expected.\n",
	     counter, result, jpegDataPtr->size);
    }
  } /* end for... */
  close(outputFile);						    /* Done writing JPEG data to disk */
  result = gettimeofday(&after, (struct timezone*)NULL);	    /* Get the current time after all this work has been done */
  before.tv_usec += (before.tv_sec * 1000000);			    /* Convert this timeval to all microseconds */
  after.tv_usec += (after.tv_sec * 1000000);			    /* Do the same for this timeval */
  totalTime.tv_usec = after.tv_usec - before.tv_usec;		    /* Determine the total time taken for the above compressions */
  averageDiskBandwidth = (int)((float)numberBytesWritten /
			       (float)(totalTime.tv_usec / 1000000));
  timeForOneGrab.tv_usec = (totalTime.tv_usec / totalNumberOfGrabs);
  grabsPerSecond = (1000000 / timeForOneGrab.tv_usec);
  timeForOneGrab.tv_sec = (timeForOneGrab.tv_usec / 1000000);	    /* Convert this timeval back into seconds and microseconds */
  timeForOneGrab.tv_usec = (timeForOneGrab.tv_usec % 1000000);
  totalTime.tv_sec = (totalTime.tv_usec / 1000000);		    /* Convert this timeval back into seconds and microseconds */
  totalTime.tv_usec = (totalTime.tv_usec % 1000000);
  
  printf("Number of grabs per second:\t\t\t%d\n", grabsPerSecond);
  printf("Average compression ratio:\t\t\t%d:1\n",
	 (myRect.width * myRect.height * 3 * totalNumberOfGrabs) /
	 numberBytesWritten);
  printf("Average size of each frame:\t\t\t%d bytes.\n",
	 numberBytesWritten / totalNumberOfGrabs);
  printf("Average disk bandwidth:\t\t\t\t%d bytes per second.\n",
	 averageDiskBandwidth);
  printf("------------------------------------------------------------------------\n");
  printf("Total time taken for %d grabs:\t\t\t%d.%d seconds.\n",
	 totalNumberOfGrabs,
	 totalTime.tv_sec,
	 totalTime.tv_usec);
  printf("Average time taken for 1 grab:\t\t\t%d.%d seconds.\n",
	 timeForOneGrab.tv_sec,
	 timeForOneGrab.tv_usec);
  printf("Total number of bytes written to disk:\t\t%d bytes.\n", numberBytesWritten);
  printf("Smallest frame in bytes (after compression):\t%d bytes.\n",
	 smallestFrameSize);
  printf("Largest frame in bytes (after compression):\t%d bytes.\n",
	 largestFrameSize);
  printf("Q Factor:\t\t\t\t\t%d.\n", qFactor);
  XtDestroyWidget(top);
  exit(0);
  return;
} /* end function TimerFunction */
